@@ -7,7 +7,7 @@
         - Syntax highlighing
         - Automatic placement
 
-    Copyright (c) 2012-2014 European Space Agency
+    Copyright (c) 2012-2019 European Space Agency
 
     Designed and implemented by Maxime Perrotin for the TASTE project
 
@@ -17,21 +17,17 @@
 import string
 import logging
 
-from PySide.QtCore import Qt, QRegExp, Slot
-
-from PySide.QtGui import(QGraphicsTextItem, QGraphicsProxyWidget, QListWidget,
-                         QStringListModel, QCompleter, QListWidgetItem, QFont,
-                         QTextCursor, QSyntaxHighlighter, QTextCharFormat,
-                         QTextBlockFormat, QStringListModel)
-
-import undoCommands
+from PySide2.QtCore import *
+from PySide2.QtGui import *
+from PySide2.QtWidgets import *
+from . import undoCommands
 
 __all__ = ['EditableText']
 
 LOG = logging.getLogger(__name__)
 
 # pylint: disable=R0904
-class Completer(QGraphicsProxyWidget, object):
+class Completer(QGraphicsProxyWidget):
     ''' Class for handling text autocompletion in the SDL scene '''
     def __init__(self, parent):
         ''' Create an autocompletion list popup '''
@@ -65,7 +61,7 @@ class Completer(QGraphicsProxyWidget, object):
         self._completer.setCompletionPrefix(completion_prefix)
         self.widget().clear()
         count = self._completer.completionCount()
-        for i in xrange(count):
+        for i in range(count):
             self._completer.setCurrentRow(i)
             self.widget().addItem(self._completer.currentCompletion())
         self.prepareGeometryChange()
@@ -93,7 +89,7 @@ class Completer(QGraphicsProxyWidget, object):
 
 
 # pylint: disable=R0904
-class Highlighter(QSyntaxHighlighter, object):
+class Highlighter(QSyntaxHighlighter):
     ''' Class for handling syntax highlighting in editable text '''
     def __init__(self, parent, blackbold_patterns, redbold_patterns):
         ''' Define highlighting rules - inputs = lists of patterns '''
@@ -132,20 +128,21 @@ class Highlighter(QSyntaxHighlighter, object):
 
 
 # pylint: disable=R0902
-class EditableText(QGraphicsTextItem, object):
+class EditableText(QGraphicsTextItem):
     '''
         Editable text area inside symbols
         Includes autocompletion when parent item needs it
     '''
     default_cursor = Qt.IBeamCursor
     hasParent = False
+    word_under_cursor = Signal(str)
 
     def __init__(self, parent, text='...', hyperlink=None):
         super(EditableText, self).__init__(parent)
+        self.parent = parent
         self.setFont(QFont('Ubuntu', 10))
         self.completer = Completer(self)
-        self.completer.widget().itemActivated.connect(
-                self.completion_selected)
+        self.completer.widget().itemActivated.connect(self.completion_selected)
         self.hyperlink = hyperlink
         self.setOpenExternalLinks(True)
         if hyperlink:
@@ -162,11 +159,12 @@ class EditableText(QGraphicsTextItem, object):
         self.highlighter = Highlighter(
                 self.document(), parent.blackbold, parent.redbold)
         self.completion_prefix = ''
-        self.set_textbox_position()
+        self.old_word = ''  # needed to detect change of word under cursor
+        #self.set_textbox_position()
         self.set_text_alignment()
         # Increase the Z value of the text area so that the autocompleter
         # always appear on top of text's siblings (parents's followers)
-        self.setZValue(1)
+        self.setZValue(self.zValue() + 1)
         # context is used for advanced autocompletion
         self.context = ''
         # Set cursor when mouse goes over the text
@@ -174,10 +172,11 @@ class EditableText(QGraphicsTextItem, object):
         # Activate cache mode to boost rendering by calling paint less often
         # Removed - does not render text properly (eats up the right part)
         # self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
+        self.force_focus = False
 
     def set_text_alignment(self):
         ''' Apply the required text alignment within the text box '''
-        alignment = self.parentItem().text_alignment
+        alignment = self.parent.text_alignment
         self.setTextWidth(self.boundingRect().width())
         fmt = QTextBlockFormat()
         fmt.setAlignment(alignment)
@@ -189,40 +188,42 @@ class EditableText(QGraphicsTextItem, object):
 
     def set_textbox_position(self):
         ''' Compute the textbox position '''
-        parent_rect = self.parentItem().boundingRect()
+        parent_rect = self.parent.boundingRect()
         rect = self.boundingRect()
         # Use parent symbol alignment requirement
-        # Does not support right nor bottom alignment
-        alignment = self.parentItem().textbox_alignment
+        # Does not support right alignment (just add it when needed)
+        alignment = self.parent.textbox_alignment
         rect_center = parent_rect.center() - rect.center()
         if alignment & Qt.AlignLeft:
             x_pos = 0
         elif alignment & Qt.AlignHCenter:
             x_pos = rect_center.x()
+        else:
+            x_pos = 0
         if alignment & Qt.AlignTop:
             y_pos = 0
         elif alignment & Qt.AlignVCenter:
             y_pos = rect_center.y()
+        elif alignment & Qt.AlignBottom:
+            y_pos = parent_rect.height()
+        else:
+            y_pos = 0
         self.setPos(x_pos, y_pos)
 
-    def paint(self, painter, _, ___):
-        ''' Place the textbox in the parent symbol and draw it '''
-        self.set_textbox_position()
-        super(EditableText, self).paint(painter, _, ___)
-
     def try_resize(self):
         '''
             If needed, request a resizing of the parent item
             (when text size expands)
         '''
-        if self.parentItem().auto_expand:
+        if self.parent.auto_expand:
             self.setTextWidth(-1)
-            parent_rect = self.parentItem().boundingRect()
+            parent_rect = self.parent.boundingRect()
             rect = self.boundingRect()
-            if rect.width() + 30 > parent_rect.width():
-                parent_rect.setWidth(rect.width() + 30)
+            if rect.width() + 15 > parent_rect.width():
+                parent_rect.setWidth(rect.width() + 15)
             parent_rect.setHeight(max(rect.height(), parent_rect.height()))
-            self.parentItem().resize_item(parent_rect)
+            self.parent.resize_item(parent_rect)
+            self.set_textbox_position()
 
     @Slot(QListWidgetItem)
     def completion_selected(self, item):
@@ -262,7 +263,7 @@ class EditableText(QGraphicsTextItem, object):
         try:
             # If not the first line of the text, Qt adds u+2029 as 1st char
             line = cursor.selectedText().replace(u'\u2029', '')
-            if line[pos] in string.ascii_letters + '!' + '.' + '_':
+            if line[pos] in string.ascii_letters + '!' + '.' + '_' + ':':
                 self.context = line[slice(0, pos + 1)].split()[-1]
             else:
                 self.context = ''
@@ -293,6 +294,11 @@ class EditableText(QGraphicsTextItem, object):
         text_cursor.select(QTextCursor.WordUnderCursor)
         self.completion_prefix = text_cursor.selectedText()
 
+        # "self.completion_prefix" is the complete word under the cursor
+        if self.completion_prefix != self.old_word:
+            self.word_under_cursor.emit(self.completion_prefix)
+            self.old_word = self.completion_prefix
+
         self.context_completion_list(force=(event.key()==Qt.Key_F8))
 
         completion_count = self.completer.set_completion_prefix(
@@ -315,9 +321,10 @@ class EditableText(QGraphicsTextItem, object):
 
             self.completer.setPos(pos_x, pos_y)
             self.completer.show()
+            self.force_focus = True   # avoid unwanted syntax checks
             # Make sure parent item has higher visibility than its siblings
             # (useful in decision branches)
-            self.parentItem().setZValue(1)
+            self.parent.setZValue(1)
             self.completer.setFocusProxy(self)
             self.setTabChangesFocus(True)
         else:
@@ -327,6 +334,16 @@ class EditableText(QGraphicsTextItem, object):
             self.setFocus()
         self.completer_has_focus = False
 
+    def mousePressEvent(self, event):
+        '''
+            If the completer box is active while the user clicks on another
+            area of the text box, make it disappear first
+        '''
+        if self.completer.isVisible():
+            self.completer.hide()
+            self.completer.resize(0, 0)
+        super(EditableText, self).mousePressEvent(event)
+
     # pylint: disable=C0103
     def focusOutEvent(self, event):
         '''
@@ -339,10 +356,18 @@ class EditableText(QGraphicsTextItem, object):
         if self.completer and not self.completer_has_focus:
             self.completer.hide()
             self.completer.resize(0, 0)
+        if self.force_focus:
+            # when user double-clicks on the Completer, it may be out of
+            # the editable text. It is not right to leave the focus in that
+            # case, as this would generate a syntax check while in fact
+            # user is not done editing text
+            self.setFocus()
+            self.force_focus = False
+            return
         if not self.completer or not self.completer.isVisible():
             # Trigger a select - side effect makes the toolbar update
             try:
-                self.parentItem().select(True)
+                self.parent.select(True)
             except AttributeError:
                 # Some parents may not be selectable (e.g. Signalroute)
                 pass
@@ -352,40 +377,53 @@ class EditableText(QGraphicsTextItem, object):
                 text_cursor.clearSelection()
                 self.setTextCursor(text_cursor)
             # If something has changed, check syntax and create undo command
-            if(self.oldSize != self.parentItem().boundingRect() or
-                                                self.oldText != unicode(self)):
+            if(self.oldSize != self.parent.boundingRect()
+                    or self.parent.syntax_error or self.oldText != str(self)):
                 # Call syntax checker from item containing the text (if any)
-                self.scene().check_syntax(self.parentItem())
+                if self.scene().check_syntax(self.parent):
+                    # Keep focus
+                    self.setFocus()
+                    return
                 # Update class completion list
                 self.scene().update_completion_list(self.parentItem())
                 # Create undo command, including possible CAM
                 with undoCommands.UndoMacro(self.scene().undo_stack, 'Text'):
                     undo_cmd = undoCommands.ResizeSymbol(
-                                          self.parentItem(), self.oldSize,
-                                          self.parentItem().boundingRect())
+                                          self.parent, self.oldSize,
+                                          self.parent.boundingRect())
                     self.scene().undo_stack.push(undo_cmd)
                     try:
-                        self.parentItem().cam(self.parentItem().pos(),
-                                              self.parentItem().pos())
+                        self.parent.cam(self.parent.pos(),
+                                              self.parent.pos())
                     except AttributeError:
                         # Some parents may not have CAM function (e.g. Channel)
                         pass
 
                     undo_cmd = undoCommands.ReplaceText(self, self.oldText,
-                                                        unicode(self))
+                                                        str(self))
                     self.scene().undo_stack.push(undo_cmd)
         self.set_text_alignment()
+        # Reset Z-Values that were increased when getting focus
+        top_level = self.parent.top_level()
+        top_level.setZValue(top_level.zValue() - 1)
+        self.parent.setZValue(self.parent.zValue() - 1)
         super(EditableText, self).focusOutEvent(event)
 
     # pylint: disable=C0103
     def focusInEvent(self, event):
         ''' When user starts editing text, save previous state for Undo '''
         super(EditableText, self).focusInEvent(event)
+        # Change the Z-value of items to make sure the
+        # completer is always be on top of other symbols
+        top_level = self.parent.top_level()
+        top_level.setZValue(top_level.zValue() + 1)
+        self.parent.setZValue(self.parent.zValue() + 1)
+
         # Trigger a select - side effect makes the toolbar update
         try:
-            self.parentItem().select(True)
+            self.parent.select(True)
         except AttributeError:
-            # Some parents may not be selectable (e.g. Signalroute)
+            # Some parents may not be selectable
             pass
         # Update completer list of keywords
         self.context = ''
@@ -395,15 +433,10 @@ class EditableText(QGraphicsTextItem, object):
         # Set width to auto-expand, and disables alignment, while editing:
         self.setTextWidth(-1)
         if not self.editing:
-            self.oldText = unicode(self)
-            self.oldSize = self.parentItem().boundingRect()
+            self.oldText = str(self)
+            self.oldSize = self.parent.boundingRect()
             self.editing = True
 
     def __str__(self):
         ''' Print the text inside the symbol '''
-        raise TypeError('Use UNICODE, not string!')
-
-    def __unicode__(self):
         return self.toPlainText()
-
-
